from termcolor import cprint, colored

_DEBUG = False

_COLORS = ['light_red', 'green', 'yellow', 'blue', 'light_magenta']


def set_debug(enabled: bool) -> None:
    global _DEBUG
    _DEBUG = enabled


def message(author: str, text: str) -> None:
    author = author.upper()
    idx = sum(ord(char) for char in author) % len(_COLORS)
    author_color = _COLORS[idx]
    cprint(colored(author, author_color, None, []))
    print(text)
    print('')


def info(text: str) -> None:
    cprint(text, 'cyan')


def debug(text: str, indent='    ') -> None:
    if _DEBUG:
        indented = '\n'.join([f'{indent}{x}' for x in text.splitlines()])
        cprint(indented, 'grey')



import argparse
import readline  # noqa

from golemgpt.settings import Settings
from golemgpt.memory.localfiles import LocalFilesMemory
from golemgpt.golems.roles.director import DirectorGolem
from golemgpt.utils import chdir, console, genkey


def main():
    settings = Settings()
    console.set_debug(settings.GOLEM_DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-j', '--job-key', metavar='JOB_KEY',
        type=str, default=genkey(),
        required=False
    )
    args = parser.parse_args()

    job_key = args.job_key
    console.info(f"Job key: {job_key}")

    workdir = settings.WORKDIR.absolute()
    outdir = workdir / job_key / 'output'
    outdir.mkdir(parents=True, exist_ok=True)

    memory = LocalFilesMemory(workdir)
    memory.load(job_key)
    goals = memory.goals

    while not goals:
        goal = input("Enter a goal for the Golem-GPT:\n?> ").strip()
        if goal:
            goals.append(goal)

    director = DirectorGolem(
        goals=goals,
        job_key=job_key,
        memory=memory,
        settings=settings,
    )
    with chdir(outdir):
        director.start_job()


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass



from re import compile as re_compile, DOTALL
from typing import List
from json import loads as json_loads, JSONDecodeError
from golemgpt.utils import console
from golemgpt.utils.exceptions import ParseActionsError
from .prompts import (
    INSTRUCTIONAL_PROMPT, OUTPUT_FORMAT_PROMPT, KNOWN_ACTIONS_PROMT,
    CHECK_COMPLETION_PROMPT, INITIALIZER_REPLY_EXAMPLE,
)


class GeneralLexicon:
    role_model_prompt: str = ""

    # Not-a-JSON naive preamble regex:
    naive_preamble_re = re_compile(r'([^\[\{]*)', DOTALL)

    def goal_prompt(self, goal: str) -> str:
        goal = f"{goal.strip().rstrip('.')}. Then finish the job."
        return f"The goal is: {goal}"

    def yesno_prompt(self, question: str) -> str:
        return f"Answer 'yes' or 'no'. {question}"

    def action_result_prompt(self, action: str, result: str) -> str:
        return f'Completed "{action}" with result:\n{result}'

    def remind_format_prompt(self) -> str:
        return 'If finished, just use a single "finish_job" action.'

    def guess_finish_prompt(self, reply: str) -> str:
        preamble = self.lexicon.find_preamble(reply)
        statement = preamble or reply
        return (
            "Does the following mean current job is finished "
            "(optional ask to start a new one)?"
            f"\n\n{statement}"
        )

    def initializer_prompt(self) -> str:
        return (
            f"{self.role_model_prompt}\n\n"
            f"{INSTRUCTIONAL_PROMPT}\n\n"
            f"{OUTPUT_FORMAT_PROMPT}\n\n"
            f"{KNOWN_ACTIONS_PROMT}\n\n"
            f"{CHECK_COMPLETION_PROMPT}\n\n"
        ).strip()

    def initializer_history(self) -> list:
        prompt = self.initializer_prompt()
        reply = INITIALIZER_REPLY_EXAMPLE
        return [
            {'role': 'user', 'content': prompt},
            {'role': 'assistant', 'content': reply},
        ]

    def find_preamble(self, reply: str) -> str:
        preamble = self.naive_preamble_re.search(reply)
        if preamble:
            return preamble.group()
        return ''

    def parse_reply(self, reply: str) -> List[dict]:
        """Parse the reply into an action plan."""
        console.debug(f"Parse plan:\n{reply}\n")

        # Remove not-a-JSON preamble, sometimes JSON is
        # prepended with some text, like "Here is your JSON:"
        preamble = self.find_preamble(reply)
        if preamble:
            reply = reply[len(preamble):]
            console.debug(f"Parse plan (trunc.):\n{reply}\n")

        reply = reply.strip()
        if not reply:
            raise ParseActionsError("JSON not found", reply)

        try:
            parsed = json_loads(reply)
        except JSONDecodeError as exc:
            raise ParseActionsError(exc.msg, reply) from exc

        if isinstance(parsed, dict):
            parsed = [parsed]

        return parsed



from golemgpt.utils import console
from golemgpt.utils.exceptions import AlignAcionsError
from .base import BaseCodex, BaseCodexLexicon


class ReasonableCodexLexicon(BaseCodexLexicon):
    def align_actions_prompt(self, action_plan: list) -> str:
        question = f"In this action plan unharmful? Action:\n{action_plan}"
        return f"Answer just 'yes' or 'no'. {question}"


class ReasonableCodex(BaseCodex):
    check_actions_depth = 3
    name = 'reasonable-codex'
    lexicon = ReasonableCodexLexicon()

    def align_actions(self, action_plan: list) -> str:
        check_actions = action_plan[:self.check_actions_depth]
        prompt = self.lexicon.align_actions_prompt(check_actions)
        console.message(self.name, prompt)
        if self.cognitron.ask_yesno(prompt):
            return True
        raise AlignAcionsError()



from golemgpt.memory import BaseMemory
from golemgpt.settings import Settings


class BaseCognitron:
    DEFAULT_NAME = 'Cognitron'

    def __init__(
        self, settings: Settings, memory: BaseMemory, name: str = ''
    ) -> None:
        self.settings = settings
        self.memory = memory
        self.name = name or self.DEFAULT_NAME

    def communicate(self, message: str, **options) -> str:
        """Communicate with the Cognitron and return the reply."""
        raise NotImplementedError()

    def ask_yesno(self, message: str, **options) -> str:
        """Ask the Cognitron a yes/no question and return the reply."""
        yesno = self.communicate(message)
        return yesno.lower().startswith('yes')

    def get_last_message(self) -> str:
        """Get the last message from the dialog with the Cognitron."""
        return self.memory.messages[-1]['content']



def delegate_job_action(
    goal: str, role: str, in_files: list, out_files: list, **kwargs
) -> str:
    """Delegate a job and return its content."""
    return "Rejected, this action is not implemented."



from typing import Any, Optional, Tuple
from golemgpt.settings import Settings
from golemgpt.utils.exceptions import UnknownAction
from golemgpt.actions import ALL_KNOWN_ACTIONS


class JustDoRunner:
    """A naive runner that just performs the specified action."""

    def __init__(
        self, settings: Settings, known_actions: Optional[dict] = None
    ) -> None:
        self.settings = settings
        self.known_actions = known_actions or ALL_KNOWN_ACTIONS

    def __str__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __call__(self, action_item: dict, golem: Any) -> Tuple[str, str]:
        for key in action_item:
            if key not in self.known_actions:
                raise UnknownAction(key)
            action_fn = self.known_actions[key]
            kwargs = action_item[key]
            result = action_fn(**kwargs)
            break
        return (key, result)



from pathlib import Path
from pydantic import BaseSettings


class Settings(BaseSettings):
    GOLEM_DEBUG: bool = False
    OPENAI_API_KEY: str = ''
    OPENAI_ORG_ID: str = ''
    OPENAI_MODEL: str = 'gpt-3.5-turbo-0301'
    WORKDIR: Path = Path('workdir')

    class Config:
        env_file = '.env'



from typing import List
from golemgpt.codex import BaseCodex
# from golemgpt.codex import UnawareCodex
from golemgpt.codex import ReasonableCodex
from golemgpt.lexicon import GeneralLexicon
from golemgpt.runners import JustDoRunner
from golemgpt.settings import Settings
from golemgpt.memory import BaseMemory
from golemgpt.utils import console, genkey
from golemgpt.cognitron.openai import OpenAICognitron
from golemgpt.utils.exceptions import (
    JobFinished, JobRejected, ParseActionsError, AlignAcionsError
)

DEFAULT_NAME = 'Golem-GPT'

DEFAULT_RETRY_PLAN_ATTEMPTS = 3


class GeneralGolem:
    cognitron_class = OpenAICognitron
    lexicon_class = GeneralLexicon
    codex_class = ReasonableCodex
    runner_class = JustDoRunner

    # TODO: Spawn a new Golem to make parseable action plan from the reply ?

    def __init__(
        self, *, name=DEFAULT_NAME, goals: List[str], job_key: str,
        memory: BaseMemory, settings: Settings
    ) -> None:
        self.name = name
        self.action_plan = []
        self.job_key = job_key
        self.goals = goals
        self.memory = memory
        self.settings = settings
        self.lexicon = self.lexicon_class()
        self.runner = self.runner_class(settings)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.job_key})"

    def start_job(self) -> None:
        """Start the job."""
        goals_text = '\n'.join(self.goals)
        console.info(f"Starting job: {self.job_key}")
        console.info(f"Goals:\n{goals_text}")
        #
        self.initialize()
        outcome = self.lexicon.goal_prompt(self.goals[-1])
        while True:
            try:
                if outcome:
                    self.plan_actions(outcome)
                    self.align_actions(outcome)
                outcome = self.run_action()
            except JobFinished:
                break
            except JobRejected:
                break
            except AlignAcionsError:
                # TODO: Implement a retry plan mechanism after misalignment
                break
        #
        console.info(f"Job completed: {self.job_key}")

    def initialize(self) -> None:
        """Initialize the job state from memory or from scratch."""
        console.debug(f"Syncing job state with memory: {self.job_key}")
        self.memory.load(self.job_key)

        if self.memory.is_history_empty:
            self.memory.goals = self.goals
            iniital_history = self.lexicon.initializer_history()
            for message in iniital_history:
                self.memory.messages.append(message)
        else:
            self.goals = self.memory.goals
            assert self.goals

        self.memory.save()
    
    def cognitron(self, memory: BaseMemory, **options) -> OpenAICognitron:
        """Return a Cognitron instance."""
        return self.cognitron_class(
            settings=self.settings, memory=memory, **options
        )

    def codex(self, **options) -> BaseCodex:
        """Return a Codex instance."""
        key = f'{self.job_key}.{genkey()}'
        memory = self.memory.spawn(key)
        cognitron = self.cognitron(memory, name='', **options)
        return self.codex_class(cognitron)

    def run_action(self) -> str:
        """Run the next action in the plan."""
        if not self.action_plan:
            raise JobFinished()
        action_item = self.action_plan.pop(0)
        action, result = self.runner(action_item, golem=self)
        if not result:
            return ''
        return self.lexicon.action_result_prompt(action, result)

    def plan_actions(self, prompt: str, attempt: int = 0) -> None:
        """Ask to update the plan based on the prompt."""
        console.message(self.name, prompt)
        reply = self.cognitron(self.memory).communicate(prompt)
        try:
            self.action_plan = self.lexicon.parse_reply(reply)
        except ParseActionsError:
            self.try_restore_plan(reply, attempt + 1)

    def align_actions(self, prompt: str) -> List[str]:
        """Align the action plan with the codex."""
        return self.codex().align_actions(self.action_plan)

    # TODO: Extract restore strategy to a separate class
    def try_restore_plan(self, reply: str, attempt: int = 0) -> None:
        """Try restore the plan after malformed reply."""
        # Finish if too many failed attempts:
        if attempt > DEFAULT_RETRY_PLAN_ATTEMPTS:
            raise JobFinished()

        # Ask in a helper dialog, if job is finished:
        question = self.lexicon.guess_finish_prompt(reply)
        if self.helper_yesno(question):
            raise JobFinished()

        # Try to plan again after remainder about the format:
        remainder = self.lexicon.remind_format_prompt()
        self.plan_actions(remainder, attempt + 1)

    # TODO: Extract helper dialogs to a separate class (?)
    def helper_yesno(self, question: str) -> bool:
        """Guess if the reply is a yes or no."""
        prompt = self.lexicon.guess_yesno_prompt(question)
        key = f'{self.job_key}.{genkey()}'
        memory = self.memory.spawn(key)
        cognitron = self.cognitron(memory, name='Helper')
        return cognitron.ask_yesno(prompt)




import time
import random
import string
from pathlib import Path
from .exceptions import PathRejected


def genkey() -> str:
    prefix = hex(int(time.time()))[1:]
    suffix = ''.join(random.choices(string.ascii_lowercase, k=3))
    return f'{prefix}-{suffix}'


def workpath(relpath: str) -> Path:
    cwd = Path.cwd().absolute()
    path = Path(Path.cwd() / relpath).absolute()
    if cwd in path.parents:
        return path
    raise PathRejected(relpath, f"File {relpath} is outside of working dir.")



from golemgpt.utils import console
from golemgpt.settings import Settings
from golemgpt.memory import BaseMemory
from golemgpt.utils.http import http_request
from .base import BaseCognitron


class OpenAICognitron(BaseCognitron):
    DEFAULT_NAME = 'OpenAI'
    MAX_TOKENS = 0
    TEMPERATURE = 0.1
    COMPLETIONS_URL = 'https://api.openai.com/v1/chat/completions'

    def __init__(
        self, settings: Settings, memory: BaseMemory, **options
    ) -> None:
        super().__init__(settings, memory, **options)
        self.headers = {
            'authorization': f'Bearer {settings.OPENAI_API_KEY}',
            'openai-organization': settings.OPENAI_ORG_ID,
        }
        self.model = settings.OPENAI_MODEL
        self.memory = memory

    def communicate(self, message: str, **options) -> str:
        """Communicate with the OpenAI and return the reply."""
        max_tokens = options.get('max_tokens', self.MAX_TOKENS)
        temperature = options.get('temperature', self.TEMPERATURE)
        messages = self.memory.messages.copy()
        messages.append({'role': 'user', 'content': message})
        payload = {
            'model': self.model,
            'messages': messages,
            'temperature': temperature,
        }
        if max_tokens:
            payload.update({'max_tokens': max_tokens})

        result = http_request(
            url=self.COMPLETIONS_URL, method='POST',
            headers=self.headers, json=payload
        )
        reply = result['choices'][0]['message']

        messages.append(reply)
        self.memory.messages = messages
        self.memory.save()

        reply_text = self.get_last_message()
        console.message(self.name, reply_text)
        return reply_text



from .ask_google import ask_google_action
from .ask_human_input import ask_human_input_action
from .create_python_script import create_python_script_action
from .create_shell_script import create_shell_script_action
from .delegate_job import delegate_job_action
from .explain import explain_action
from .finish_job import finish_job_action
from .get_local_date import get_local_date_action
from .get_os_details import get_os_details_action
from .http_download import http_download_action
from .read_file import read_file_action
from .reject_job import reject_job_action
from .run_script import run_script_action
from .write_file import write_file_action

__all__ = [
    'ALL_KNOWN_ACTIONS',
]

ALL_KNOWN_ACTIONS = {
    'ask_human_input': ask_human_input_action,
    'ask_google': ask_google_action,
    #
    'http_download': http_download_action,
    #
    'read_file': read_file_action,
    'write_file': write_file_action,
    #
    'get_local_date': get_local_date_action,
    'get_os_details': get_os_details_action,
    #
    'create_python_script': create_python_script_action,
    'create_shell_script': create_shell_script_action,
    'run_script': run_script_action,
    #
    'delegate_job': delegate_job_action,
    'reject_job': reject_job_action,
    'finish_job': finish_job_action,
    #
    'explain': explain_action,
}




from golemgpt.utils import console
from .base import BaseCodex


class UnawareCodex(BaseCodex):
    def align_actions(self, action_plan: list) -> str:
        console.debug(f"Naively approve actions: {action_plan}")
        return True



import subprocess
from golemgpt.utils import workpath


def run_script_action(name: str, **kwargs) -> str:
    """Run a script and return its output."""
    try:
        path = workpath(name)
    except ValueError as exc:
        return f"Rejected: {exc}."

    if not path.exists():
        return f"Rejected: script {name} does not exist."

    if name.endswith('.py'):
        process = subprocess.Popen(['python', name])
    elif name.endswith('.sh'):
        process = subprocess.Popen(['sh', name])
    else:
        return f"Rejected: script {name} is not a Bash or Python script."

    process.wait()
    is_ok = (process.returncode == 0)

    return (
        f"Script {name} executed successfully."
        if is_ok else
        f"Script {name} failed."
    )



from pathlib import Path

MAX_SIZE = 8096


def read_file_action(filename: str, **kwargs) -> str:
    """Read a file and return its content."""
    cwd = Path.cwd().absolute()
    path = Path(Path.cwd() / filename).absolute()

    if cwd in path.parents:
        relname = path.relative_to(cwd)
    else:
        return f"Rejected, file {filename} is outside of working dir."

    if not path.exists():
        return f"Rejected, file {relname} does not exist."

    file_size = path.stat().st_size
    if path.stat().st_size > MAX_SIZE:
        return f"File exists, but is too large ({file_size} bytes)."

    with path.open("r") as file:
        return file.read()



from typing import Tuple


class AlwaysAskRunner:
    """A runner that asks for confirmation before running an action."""

    def __call__(self, action: dict) -> Tuple[str, str]:
        yes = input(f"Press Y to run action: {action}")
        if yes.lower() != 'y':
            return '', ''
        super().__call__(action)



from golemgpt.cognitron import BaseCognitron


class BaseCodexLexicon:
    def align_actions(self, action_plan: list) -> str:
        raise NotImplementedError()


class BaseCodex:
    def __init__(self, cognitron: BaseCognitron):
        self.cognitron = cognitron

    def align_actions(self, action_plan: list) -> str:
        raise NotImplementedError()



from golemgpt.utils.exceptions import JobFinished


def ask_human_input_action(query: str, **kwargs):
    try:
        result = input(f"{query}\n?> ")
    except KeyboardInterrupt:
        raise JobFinished()
    return f"{query}\nAnswer is: {result}"



from .delegate_job import delegate_job_action


def create_shell_script_action(
    name: str, description: str, in_files: list, out_files: list, **kwargs
) -> str:
    """Create a Bash script and return its content."""
    return delegate_job_action(
        goal=f"Create a Bash script {name}: {description}",
        role='coder',
        in_files=in_files,
        out_files=out_files,
    )



from golemgpt.golems.general import GeneralGolem
from golemgpt.lexicon.general import GeneralLexicon

# Use the real APIs, trigger search actions to get extra information when needed.
# Autonomous bots are preferred. Ask for all prerequisite on start.
# For example, travel dates, preferences, or API credentials.

# Act as a Solutions architect and expert in APIs to create pipeline for bots
# to achieve the specified goals. Autonomous bots are preferred.

PROMPT_FOR_DIRECTOR = """
Act as a Solutions architect and expert in APIs.
""".strip()


class DirectorLexicon(GeneralLexicon):
    role_model_prompt: str = PROMPT_FOR_DIRECTOR


class DirectorGolem(GeneralGolem):
    lexicon_class = DirectorLexicon



from pydantic import BaseSettings


class Settings(BaseSettings):
    OPENAI_API_KEY: str = ''
    OPENAI_ORG_ID: str = ''
    OPENAI_IMAGE_SIZE = '1024x1024'
    HISTORY_ROOT = './history'

    class Config:
        env_file = '.env'
